#include <cstdio>
#include <cstring>
#include <cstdlib>

#include <vector>
#include <algorithm>

#include "divsufsort.h"
#include "rmq_tree.h"
#include "ms_support.h"
#include "match.h"
#include "parse.h"
#include "reader.h"
#include "endmap.h"
#include "uint40.h"
#include "utils.h"
#include "stream.h"

const int skip_threshold = 40;

void parse_phrase(unsigned char *x, int n, int i,
    int psv, int nsv, int &pos, int &len) {
  pos = -1;
  len = 0;

  if (nsv != -1 && psv != -1) {
    while (x[psv + len] == x[nsv + len])
      ++len;
    if (x[i + len] == x[psv + len]) {
      ++len;
      while (x[i + len] == x[psv + len])
        ++len;
      pos = psv;
    } else {
      while (i + len < n && x[i + len] == x[nsv + len])
        ++len;
      pos = nsv;
    }
  } else if (psv != -1) {
    while (x[psv + len] == x[i + len])
      ++len;
    pos = psv;
  } else if (nsv != -1) {
    while (i + len < n && x[nsv + len] == x[i + len])
      ++len;
    pos = nsv;
  }

  if (!len) pos = x[i];
}

long serialize(long p, int l) { return (p << 20) | l; }
void deserialize(long x, long &p, int &l) {
  static const long mask = (1L << 20) - 1;
  l = x & mask; x >>= 20; p = x;
}

void parse(
    std::string in_fname,
    std::string out_fname,
    long max_block_size) {
  #ifdef SHOWTIME
    long double start = utils::wclock();
  #endif

  long parsed = 0;   // factorized prefix length
  long nphrases = 0; // current parsing size
  long length = utils::file_size(in_fname.c_str());

  stream_writer<uint40> *writer = NULL;
  if (out_fname.length()) {
    writer = new stream_writer<uint40>(out_fname);
    writer->write(length);
  }

  file_reader fr(in_fname.c_str(), length);

  static const int endbuf_size = (1 << 20);
  long *endbuf = new long[endbuf_size];
  end_map em("endmap", (length / skip_threshold) + 1000);


  long toalloc = 29L * max_block_size + (1 << 20);
  unsigned char *memory = new unsigned char[toalloc];
  std::fill(memory, memory + toalloc, 0);

  #ifdef SHOWTIME
  fprintf(stderr, "Allocating memory: %ld bytes\n", toalloc);
  #endif

  while (parsed < length) {
    #ifdef SHOWTIME
      long double parsed_MiB = (long double)parsed / (1 << 20);
      long double elapsed = utils::wclock() - start;
      fprintf(stderr, "Parsed %.1LfMB, z=%ld, tim=%.2Lfs (%.2Lfs/MiB). Next block:\n",
          parsed_MiB, nphrases, elapsed, elapsed / parsed_MiB);
    #endif

    long block_size = std::min(max_block_size, length - parsed);

    unsigned char *free_mem_ptr = memory;
    uint40 *ms_pos = (uint40 *)free_mem_ptr; free_mem_ptr += sizeof(uint40) * block_size + 20;
    while ((free_mem_ptr - memory) % 4) ++free_mem_ptr;
    unsigned char *b = free_mem_ptr;    free_mem_ptr += block_size;
    while ((free_mem_ptr - memory) % 4) ++free_mem_ptr;
    int *ms_len = (int *)free_mem_ptr;  free_mem_ptr += 4 * block_size + 10;
    while ((free_mem_ptr - memory) % 4) ++free_mem_ptr; 
    int *SA = (int *)free_mem_ptr;      free_mem_ptr += 4 * block_size + 10;
    while ((free_mem_ptr - memory) % 4) ++free_mem_ptr;

    // Read block.
    #ifdef SHOWTIME
      fprintf(stderr, "  Reading block: ");
      long double reading_start = utils::wclock();
    #endif

    fr.read_block(b, block_size, length - parsed - block_size);

    #ifdef SHOWTIME
      long double reading_time = utils::wclock() - reading_start;
      fprintf(stderr, "%.2Lf\n", reading_time);
    #endif

    std::reverse(b, b + block_size);

    rmq_tree *rmq = NULL;
    ms_support *index = NULL;

    #ifdef SHOWTIME
      fprintf(stderr, "  Suffix array: ");
      long double sa_start = utils::wclock();
    #endif

    divsufsort(b, SA, block_size);

    #ifdef SHOWTIME
      fprintf(stderr, "\r  Suffix array: %.2Lf\n", utils::wclock() - sa_start);
    #endif

    std::fill(ms_len, ms_len + block_size, 0);
    if (parsed > 0) {
      // Compute matching statistics for the current block.

      #ifdef SHOWTIME
        fprintf(stderr, "  MS-support: ");
        long double ms_start = utils::wclock();
      #endif

      index = new ms_support(b, SA, block_size, free_mem_ptr);

      #ifdef SHOWTIME
        fprintf(stderr, "\r  MS-support: %.2Lf\n", utils::wclock() - ms_start);
      #endif


      // For each i we compute pos, lcp such that x[i..i+lcp) occurs in b,
      // lcp is maximized and x[i..i+lcp) = b[SA[pos]..SA[pos]+lcp). 

      long phrase_start;
      int phrase_length;

      if (em.isnext())
        deserialize(em.next(), phrase_start, phrase_length);
      else {
        phrase_start = -1;
        phrase_length = 0;
      }

      #ifdef SHOWTIME
        fprintf(stderr, "  Stream: \r");
        long double stream_start = utils::wclock();
      #endif

      int pos = index->dollar, lcp = block_size;
      for (long i = parsed - 1; i >= 0; --i) {
        index->extend_left(pos, lcp, fr.get(length - 1 - i));


        if (lcp && ms_len[pos] < lcp) {
          ms_len[pos] = lcp;
          ms_pos[pos] = i;
        }

        // Skipping trick. First find the candidate for enclosing phrase.
        if (i < phrase_start) {
          if (em.isnext()) deserialize(em.next(), phrase_start, phrase_length);
          else { phrase_start = -1; phrase_length = 0; }
        }

        // If it is long enough and x[i..i+lcp) falls inside, restart
        // matching statistics at the beginning of that phrase.
        if (i + lcp - 1 < phrase_start + phrase_length) {
          i = phrase_start;
          std::pair<int, int> match_pair = index->longest_prefix(i, length, &fr);
          pos = match_pair.first;
          lcp = match_pair.second;
          continue;
        }
      }

      #ifdef SHOWTIME
        long double stream_time = utils::wclock() - stream_start;
        long double streamed_MiB = (long double)parsed / (1 << 20);
        fprintf(stderr, "  Stream: 100.0%%. Time: %.2Lf (speed %.2LfMB/s)\n",
            stream_time, streamed_MiB / stream_time);
      #endif

      // Matching statistics inversion.
      for (int i = 1; i < block_size; ++i)
        if (std::min(ms_len[i - 1], index->LCP[i]) > ms_len[i]) {
          ms_len[i] = std::min(ms_len[i - 1], index->LCP[i]);
          ms_pos[i] = ms_pos[i - 1];
        }
      for (int i = block_size - 2; i >= 0; --i)
        if (std::min(ms_len[i + 1], index->LCP[i + 1]) > ms_len[i]) {
          ms_len[i] = std::min(ms_len[i + 1], index->LCP[i + 1]);
          ms_pos[i] = ms_pos[i + 1];
        }
    }

    // Parse the currect block (taking matching stats into account).
    #ifdef SHOWTIME
      fprintf(stderr, "  Parse block: ");
      long double parse_block_start = utils::wclock();
    #endif

    int *ISA = parsed ? index->LCP : (int *)ms_pos; // Space saving.
    for (int i = 0; i < block_size; ++i) ISA[SA[i]] = i;
    rmq = new rmq_tree(SA, block_size, 7);
    int endbuf_ptr = 0;
    int i = 0, len, pos32;
    long pos;

    while(i < block_size) {
      int psv = rmq->psv(ISA[i], i), nsv = rmq->nsv(ISA[i], i);
      parse_phrase(b, block_size, i, psv, nsv, pos32, len);
      pos = len ? parsed + pos32 : b[i];

      if ((ms_len[ISA[i]] > len && len) || (!len && ms_len[ISA[i]])) {
        len = ms_len[ISA[i]];
        pos = ms_pos[ISA[i]].ll();
      }

      // Do not add phrase that can might overlap block boundary.
      if (len && i + len == block_size && parsed + block_size < length) break;

      if (writer) {
        writer->write(pos);
        writer->write(len);
      }

      if (len > skip_threshold)
        endbuf[endbuf_ptr++] = serialize(parsed + i, len);
      i += std::max(1, len);
      if (endbuf_ptr == endbuf_size) {
        std::reverse(endbuf, endbuf + endbuf_size);
        em.write_block(endbuf, endbuf_size);
        endbuf_ptr = 0;
      }
      ++nphrases;
    }

    if (endbuf_ptr > 0) {
      std::reverse(endbuf, endbuf + endbuf_ptr);
      em.write_block(endbuf, endbuf_ptr);
      endbuf_ptr = 0;
    }
 
    #ifdef SHOWTIME
      long double parse_block_time = utils::wclock() - parse_block_start;
      fprintf(stderr, "\r  Parse block: %.2Lf\n", parse_block_time);
    #endif
    
    delete index;
    delete rmq;

    if (i < (block_size + 1) / 2) {
      // Read a big chunk of text -- use all available memory.

      #ifdef SHOWTIME
        long double matching_start = utils::wclock();
      #endif

      int big_block_size = std::min(length - (parsed + i), 28L * block_size);

      b = memory;
      fr.read_block(b, big_block_size, length - (parsed + i + big_block_size));
      std::reverse(b, b + big_block_size);
      
      // Run pattern matching to find the whole phrases.
      std::pair<long, int> pp = pattern_matching(parsed + i, length, b, big_block_size, fr);

      #ifdef SHOWTIME
        fprintf(stderr, "  Running pattern matching: %.2Lf\n", utils::wclock() - matching_start);
      #endif

      // Phrases longer than big_block_size
      if (pp.second == big_block_size) {}

      // Store the computed phrase and continue.
      if (writer) {
        writer->write(pp.first);
        writer->write(pp.second);
      }

      if (pp.second > skip_threshold) endbuf[endbuf_ptr++] = serialize(parsed + i, pp.second);
      if (endbuf_ptr > 0) em.write_block(endbuf, endbuf_ptr);
      ++nphrases;
      parsed += pp.second;
    }
    parsed += i;
  }

  #ifdef SHOWTIME
    long double total_time = utils::wclock() - start;
    long double length_MiB = (long double)length / (1 << 20);
    fprintf(stderr, "Total parsing time: %.2Lfs (%.2Lfs/MiB)\n", total_time,
        total_time / length_MiB);
    fprintf(stderr, "Number of phrases = %ld\n", nphrases);
  #endif

  delete[] endbuf;
  delete[] memory;

  delete writer;
}

